@@ -9,6 +9,7 @@
 #include "xs_openssl.h"
 #include "xs_time.h"
 #include "xs_mime.h"
+#include "xs_match.h"
 
 #include "snac.h"
 
@@ -203,8 +204,7 @@ xs_str *html_msg_icon(xs_str *os, const xs_dict *msg)
         int priv    = 0;
         const char *type = xs_dict_get(msg, "type");
 
-        if (strcmp(type, "Note") == 0 || strcmp(type, "Question") == 0 ||
-            strcmp(type, "Page") == 0 || strcmp(type, "Article") == 0)
+        if (xs_match(type, "Note|Question|Page|Article"))
             url = xs_dict_get(msg, "id");
 
         priv = !is_msg_public(msg);
@@ -481,6 +481,32 @@ xs_str *html_user_header(snac *snac, xs_str *s, int local)
             xs *s1   = xs_fmt("<div class=\"p-note snac-top-user-bio\">%s</div>\n", bio2);
 
             s = xs_str_cat(s, s1);
+
+            xs_dict *metadata = xs_dict_get(snac->config, "metadata");
+            if (xs_type(metadata) == XSTYPE_DICT) {
+                xs_str *k;
+                xs_str *v;
+
+                s = xs_str_cat(s, "<br><div class=\"snac-metadata\">\n");
+
+                while (xs_dict_iter(&metadata, &k, &v)) {
+                    xs *k2 = encode_html(k);
+                    xs *v2 = encode_html(v);
+                    xs *v3 = NULL;
+
+                    if (xs_startswith(v, "https:/" "/"))
+                        v3 = xs_fmt("<a href=\"%s\">%s</a>", v2, v2);
+                    else
+                        v3 = xs_dup(v2);
+
+                    xs *s1 = xs_fmt("<span class=\"snac-property-name\">%s</span>:<br>"
+                            "<span class=\"snac-property-value\">%s</span><br>\n", k2, v3);
+
+                    s = xs_str_cat(s, s1);
+                }
+
+                s = xs_str_cat(s, "</div>\n");
+            }
         }
 
         s = xs_str_cat(s, "</div>\n");
@@ -581,6 +607,14 @@ xs_str *html_top_controls(snac *snac, xs_str *s)
         "<p><input type=\"checkbox\" name=\"bot\" id=\"bot\" %s>\n"
         "<label for=\"bot\">%s</label></p>\n"
 
+        "<p><input type=\"checkbox\" name=\"private\" id=\"private\" %s>\n"
+        "<label for=\"private\">%s</label></p>\n"
+
+        "<p>%s:<br>\n"
+        "<textarea name=\"metadata\" cols=\"40\" rows=\"4\" "
+        "placeholder=\"Blog=https:/" "/example.com/my-blog\nGPG Key=1FA54\n...\">"
+        "%s</textarea></p>\n"
+
         "<p>%s:<br>\n"
         "<input type=\"password\" name=\"passwd1\" value=\"\"></p>\n"
 
@@ -628,6 +662,7 @@ xs_str *html_top_controls(snac *snac, xs_str *s)
     const char *d_dm_f_u = xs_dict_get(snac->config, "drop_dm_from_unknown");
 
     const char *bot = xs_dict_get(snac->config, "bot");
+    const char *a_private = xs_dict_get(snac->config, "private");
 
     xs *es1 = encode_html(xs_dict_get(snac->config, "name"));
     xs *es2 = encode_html(xs_dict_get(snac->config, "bio"));
@@ -636,6 +671,19 @@ xs_str *html_top_controls(snac *snac, xs_str *s)
     xs *es5 = encode_html(telegram_chat_id);
     xs *es6 = encode_html(purge_days);
 
+    xs *metadata = xs_str_new(NULL);
+    xs_dict *md = xs_dict_get(snac->config, "metadata");
+    xs_str *k;
+    xs_str *v;
+
+    while (xs_dict_iter(&md, &k, &v)) {
+        xs *kp = xs_fmt("%s=%s", k, v);
+
+        if (*metadata)
+            metadata = xs_str_cat(metadata, "\n");
+        metadata = xs_str_cat(metadata, kp);
+    }
+
     xs *s1 = xs_fmt(_tmpl,
         L("New Post..."),
         snac->actor,
@@ -686,6 +734,12 @@ xs_str *html_top_controls(snac *snac, xs_str *s)
         L("Drop direct messages from people you don't follow"),
         xs_type(bot) == XSTYPE_TRUE ? "checked" : "",
         L("This account is a bot"),
+        xs_type(a_private) == XSTYPE_TRUE ? "checked" : "",
+        L("This account is private (posts are not shown through the web)"),
+
+        L("Profile metadata (key=value pairs in each line)"),
+        metadata,
+
         L("New password"),
         L("Repeat new password"),
         L("Update user info")
@@ -760,8 +814,10 @@ xs_str *build_mentions(snac *snac, const xs_dict *msg)
 
 xs_str *html_entry_controls(snac *snac, xs_str *os, const xs_dict *msg, const char *md5)
 {
-    char *id    = xs_dict_get(msg, "id");
-    char *actor = xs_dict_get(msg, "attributedTo");
+    const char *id    = xs_dict_get(msg, "id");
+    const char *actor = xs_dict_get(msg, "attributedTo");
+    const char *group = xs_dict_get(msg, "audience");
+
     xs *likes   = object_likes(id);
     xs *boosts  = object_announces(id);
 
@@ -774,10 +830,11 @@ xs_str *html_entry_controls(snac *snac, xs_str *os, const xs_dict *msg, const ch
             "<form autocomplete=\"off\" method=\"post\" action=\"%s/admin/action\">\n"
             "<input type=\"hidden\" name=\"id\" value=\"%s\">\n"
             "<input type=\"hidden\" name=\"actor\" value=\"%s\">\n"
+            "<input type=\"hidden\" name=\"group\" value=\"%s\">\n"
             "<input type=\"hidden\" name=\"redir\" value=\"%s_entry\">\n"
             "\n",
 
-            snac->actor, id, actor, md5
+            snac->actor, id, actor, group ? group : "", md5
         );
 
         s = xs_str_cat(s, s1);
@@ -812,6 +869,17 @@ xs_str *html_entry_controls(snac *snac, xs_str *os, const xs_dict *msg, const ch
             s = html_button(s, "follow", L("Follow"), L("Start following this user's activity"));
         }
 
+        if (!xs_is_null(group)) {
+            if (following_check(snac, group)) {
+                s = html_button(s, "unfollow", L("Unfollow Group"),
+                        L("Stop following this group or channel"));
+            }
+            else {
+                s = html_button(s, "follow", L("Follow Group"),
+                        L("Start following this group or channel"));
+            }
+        }
+
         s = html_button(s, "mute", L("MUTE"),
             L("Block any activity from this user forever"));
     }
@@ -973,8 +1041,7 @@ xs_str *html_entry(snac *user, xs_str *os, const xs_dict *msg, int local,
         return xs_str_cat(os, s);
     }
     else
-    if (strcmp(type, "Note") != 0 && strcmp(type, "Question") != 0 &&
-        strcmp(type, "Page") != 0 && strcmp(type, "Article") != 0) {
+    if (!xs_match(type, "Note|Question|Page|Article")) {
         /* skip oddities */
         return os;
     }
@@ -1325,13 +1392,10 @@ xs_str *html_entry(snac *user, xs_str *os, const xs_dict *msg, int local,
                 continue;
 
             /* infer MIME type from non-specific attachments */
-            if (xs_list_len(attach) < 2 &&
-                (strcmp(t, "Link") == 0 || strcmp(t, "Document") == 0)) {
+            if (xs_list_len(attach) < 2 && xs_match(t, "Link|Document")) {
                 const char *mt = xs_mime_by_ext(url);
 
-                if (xs_startswith(mt, "image/") ||
-                    xs_startswith(mt, "audio/") ||
-                    xs_startswith(mt, "video/"))
+                if (xs_match(mt, "image/*|audio/*|video/*")) /* */
                     t = mt;
             }
 
@@ -1501,7 +1565,7 @@ xs_str *html_timeline(snac *user, const xs_list *list, int local, int skip, int
         xs *msg = NULL;
         int status;
 
-        if (user)
+        if (user && !is_pinned_by_md5(user, v))
             status = timeline_get_by_md5(user, v, &msg);
         else
             status = object_get_by_md5(v, &msg);
@@ -1509,12 +1573,31 @@ xs_str *html_timeline(snac *user, const xs_list *list, int local, int skip, int
         if (!valid_status(status))
             continue;
 
+        /* if it's an instance page, discard private users */
+        if (user == NULL) {
+            const char *atto = xs_dict_get(msg, "attributedTo");
+            xs *l = xs_split(atto, "/");
+            const char *uid = xs_list_get(l, -1);
+            snac user;
+            int skip = 1;
+
+            if (uid && user_open(&user, uid)) {
+                if (xs_type(xs_dict_get(user.config, "private")) != XSTYPE_TRUE)
+                    skip = 0;
+
+                user_free(&user);
+            }
+
+            if (skip)
+                continue;
+        }
+
         s = html_entry(user, s, msg, local, 0, v, user ? 0 : 1);
     }
 
     s = xs_str_cat(s, "</div>\n");
 
-    if (user && local) {
+    if (list && user && local) {
         xs *s1 = xs_fmt(
             "<div class=\"snac-history\">\n"
             "<p class=\"snac-history-title\">%s</p><ul>\n",
@@ -1884,6 +1967,12 @@ int html_get_handler(const xs_dict *req, const char *q_path,
     if (p_path == NULL) { /** public timeline **/
         xs *h = xs_str_localtime(0, "%Y-%m.html");
 
+        if (xs_type(xs_dict_get(snac.config, "private")) == XSTYPE_TRUE) {
+            *body = html_timeline(&snac, NULL, 1, 0, 0, 0);
+            *b_size = strlen(*body);
+            status  = 200;
+        }
+        else
         if (cache && history_mtime(&snac, h) > timeline_mtime(&snac)) {
             snac_debug(&snac, 1, xs_fmt("serving cached local timeline"));
 
@@ -1964,6 +2053,9 @@ int html_get_handler(const xs_dict *req, const char *q_path,
     }
     else
     if (xs_startswith(p_path, "p/")) { /** a timeline with just one entry **/
+        if (xs_type(xs_dict_get(snac.config, "private")) == XSTYPE_TRUE)
+            return 403;
+
         xs *id  = xs_fmt("%s/%s", snac.actor, p_path);
         xs *msg = NULL;
 
@@ -1996,6 +2088,9 @@ int html_get_handler(const xs_dict *req, const char *q_path,
     }
     else
     if (xs_startswith(p_path, "h/")) { /** an entry from the history **/
+        if (xs_type(xs_dict_get(snac.config, "private")) == XSTYPE_TRUE)
+            return 403;
+
         xs *l    = xs_split(p_path, "/");
         char *id = xs_list_get(l, 1);
 
@@ -2012,15 +2107,18 @@ int html_get_handler(const xs_dict *req, const char *q_path,
     }
     else
     if (strcmp(p_path, ".rss") == 0) { /** public timeline in RSS format **/
+        if (xs_type(xs_dict_get(snac.config, "private")) == XSTYPE_TRUE)
+            return 403;
+
         xs_str *rss;
         xs *elems = timeline_simple_list(&snac, "public", 0, 20);
         xs *bio   = not_really_markdown(xs_dict_get(snac.config, "bio"), NULL);
         char *p, *v;
 
-        xs *es1 = encode_html(xs_dict_get(snac.config, "name"));
-        xs *es2 = encode_html(snac.uid);
-        xs *es3 = encode_html(xs_dict_get(srv_config, "host"));
-        xs *es4 = encode_html(bio);
+        xs *es1 = encode_html_strict(xs_dict_get(snac.config, "name"));
+        xs *es2 = encode_html_strict(snac.uid);
+        xs *es3 = encode_html_strict(xs_dict_get(srv_config, "host"));
+        xs *es4 = encode_html_strict(bio);
         rss = xs_fmt(
             "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
             "<rss version=\"0.91\">\n"
@@ -2048,7 +2146,7 @@ int html_get_handler(const xs_dict *req, const char *q_path,
             if (!xs_startswith(id, snac.actor))
                 continue;
 
-            xs *content = sanitize(xs_dict_get(msg, "content"));
+            xs *content = encode_html_strict(xs_dict_get(msg, "content"));
 
             // We SHOULD only use sanitized one for description.
             // So, only encode for feed title, while the description just keep it sanitized as is.
@@ -2057,7 +2155,7 @@ int html_get_handler(const xs_dict *req, const char *q_path,
             xs *title   = xs_str_new(NULL);
             int i;
 
-            for (i = 0; es_title[i] && es_title[i] != '\n' && i < 50; i++)
+            for (i = 0; es_title[i] && es_title[i] != '\n' && es_title[i] != '&' && i < 50; i++)
                 title = xs_append_m(title, &es_title[i], 1);
 
             xs *s = xs_fmt(
@@ -2268,6 +2366,7 @@ int html_post_handler(const xs_dict *req, const char *q_path,
         char *id     = xs_dict_get(p_vars, "id");
         char *actor  = xs_dict_get(p_vars, "actor");
         char *action = xs_dict_get(p_vars, "action");
+        char *group  = xs_dict_get(p_vars, "group");
 
         if (action == NULL)
             return 404;
@@ -2344,6 +2443,36 @@ int html_post_handler(const xs_dict *req, const char *q_path,
                 snac_log(&snac, xs_fmt("actor is not being followed %s", actor));
         }
         else
+        if (strcmp(action, L("Follow Group")) == 0) { /** **/
+            xs *msg = msg_follow(&snac, group);
+
+            if (msg != NULL) {
+                /* reload the group from the message, in may be different */
+                group = xs_dict_get(msg, "object");
+
+                following_add(&snac, group, msg);
+
+                enqueue_output_by_actor(&snac, msg, group, 0);
+            }
+        }
+        else
+        if (strcmp(action, L("Unfollow Group")) == 0) { /** **/
+            /* get the following object */
+            xs *object = NULL;
+
+            if (valid_status(following_get(&snac, group, &object))) {
+                xs *msg = msg_undo(&snac, xs_dict_get(object, "object"));
+
+                following_del(&snac, group);
+
+                enqueue_output_by_actor(&snac, msg, group, 0);
+
+                snac_log(&snac, xs_fmt("unfollowed group %s", group));
+            }
+            else
+                snac_log(&snac, xs_fmt("actor is not being followed %s", actor));
+        }
+        else
         if (strcmp(action, L("Delete")) == 0) { /** **/
             char *actor_form = xs_dict_get(p_vars, "actor-form");
             if (actor_form != NULL) {
@@ -2422,6 +2551,29 @@ int html_post_handler(const xs_dict *req, const char *q_path,
             snac.config = xs_dict_set(snac.config, "bot", xs_stock_true);
         else
             snac.config = xs_dict_set(snac.config, "bot", xs_stock_false);
+        if ((v = xs_dict_get(p_vars, "private")) != NULL && strcmp(v, "on") == 0)
+            snac.config = xs_dict_set(snac.config, "private", xs_stock_true);
+        else
+            snac.config = xs_dict_set(snac.config, "private", xs_stock_false);
+        if ((v = xs_dict_get(p_vars, "metadata")) != NULL) {
+            /* split the metadata and store it as a dict */
+            xs_dict *md = xs_dict_new();
+            xs *l = xs_split(v, "\n");
+            xs_list *p = l;
+            xs_str *kp;
+
+            while (xs_list_iter(&p, &kp)) {
+                xs *kpl = xs_split_n(kp, "=", 1);
+                if (xs_list_len(kpl) == 2) {
+                    xs *k2 = xs_strip_i(xs_dup(xs_list_get(kpl, 0)));
+                    xs *v2 = xs_strip_i(xs_dup(xs_list_get(kpl, 1)));
+
+                    md = xs_dict_set(md, k2, v2);
+                }
+            }
+
+            snac.config = xs_dict_set(snac.config, "metadata", md);
+        }
 
         /* uploads */
         const char *uploads[] = { "avatar", "header", NULL };
