@@ -28,9 +28,6 @@ typedef enum {
 } xstype;
 
 
-/* dynamic strings */
-typedef char d_char;
-
 /* types */
 typedef char xs_val;
 typedef char xs_str;
@@ -62,6 +59,7 @@ xs_val *xs_insert_m(xs_val *data, int offset, const char *mem, int size);
 #define xs_append_m(data, mem, size) xs_insert_m(data, xs_size(data) - 1, mem, size)
 
 xs_str *xs_str_new(const char *str);
+xs_str *xs_str_new_sz(const char *mem, int sz);
 xs_str *xs_str_wrap_i(const char *prefix, xs_str *str, const char *suffix);
 #define xs_str_prepend_i(str, prefix) xs_str_wrap_i(prefix, str, NULL)
 #define xs_str_cat(str, suffix) xs_str_wrap_i(NULL, str, suffix)
@@ -75,6 +73,8 @@ int xs_starts_and_ends(const char *prefix, const char *str, const char *suffix);
 #define xs_startswith(str, prefix) xs_starts_and_ends(prefix, str, NULL)
 #define xs_endswith(str, suffix) xs_starts_and_ends(NULL, str, suffix)
 xs_str *xs_crop_i(xs_str *str, int start, int end);
+xs_str *xs_lstrip_chars_i(xs_str *str, const char *chars);
+xs_str *xs_rstrip_chars_i(xs_str *str, const char *chars);
 xs_str *xs_strip_chars_i(xs_str *str, const char *chars);
 #define xs_strip_i(str) xs_strip_chars_i(str, " \r\n\t\v\f")
 xs_str *xs_tolower_i(xs_str *str);
@@ -115,7 +115,7 @@ const char *xs_number_str(const xs_number *v);
 
 xs_data *xs_data_new(const void *data, int size);
 int xs_data_size(const xs_data *value);
-void xs_data_get(const xs_data *value, void *data);
+void xs_data_get(void *data, const xs_data *value);
 
 void *xs_memmem(const char *haystack, int h_size, const char *needle, int n_size);
 
@@ -157,7 +157,7 @@ xs_val xs_stock_dict[]  = { XSTYPE_DICT, 0, 0, 5, XSTYPE_EOM };
 
 void *_xs_realloc(void *ptr, size_t size, const char *file, int line, const char *func)
 {
-    d_char *ndata = realloc(ptr, size);
+    xs_val *ndata = realloc(ptr, size);
 
     if (ndata == NULL) {
         fprintf(stderr, "**OUT OF MEMORY**\n");
@@ -365,19 +365,21 @@ xs_val *xs_expand(xs_val *data, int offset, int size)
 /* opens a hole in data */
 {
     int sz = xs_size(data);
+    int n;
+
+    sz += size;
 
     /* open room */
-    if (sz == 0 || _xs_blk_size(sz) != _xs_blk_size(sz + size))
-        data = xs_realloc(data, _xs_blk_size(sz + size));
+    data = xs_realloc(data, _xs_blk_size(sz));
 
     /* move up the rest of the data */
-    if (data != NULL)
-        memmove(data + offset + size, data + offset, sz - offset);
+    for (n = sz - 1; n >= offset + size; n--)
+        data[n] = data[n - size];
 
     if (xs_type(data) == XSTYPE_LIST ||
         xs_type(data) == XSTYPE_DICT ||
         xs_type(data) == XSTYPE_DATA)
-        _xs_put_24b(data + 1, sz + size);
+        _xs_put_24b(data + 1, sz);
 
     return data;
 }
@@ -427,6 +429,17 @@ xs_str *xs_str_new(const char *str)
 }
 
 
+xs_str *xs_str_new_sz(const char *mem, int sz)
+/* creates a new string from a memory block, adding an asciiz */
+{
+    xs_str *s = xs_realloc(NULL, _xs_blk_size(sz + 1));
+    memcpy(s, mem, sz);
+    s[sz] = '\0';
+
+    return s;
+}
+
+
 xs_str *xs_str_wrap_i(const char *prefix, xs_str *str, const char *suffix)
 /* wraps str with prefix and suffix */
 {
@@ -530,7 +543,7 @@ int xs_starts_and_ends(const char *prefix, const char *str, const char *suffix)
 
 
 xs_str *xs_crop_i(xs_str *str, int start, int end)
-/* crops the d_char to be only from start to end */
+/* crops the string to be only from start to end */
 {
     XS_ASSERT_TYPE(str, XSTYPE_STRING);
 
@@ -549,29 +562,39 @@ xs_str *xs_crop_i(xs_str *str, int start, int end)
 }
 
 
-xs_str *xs_strip_chars_i(xs_str *str, const char *chars)
-/* strips the string of chars from the start and the end */
+xs_str *xs_lstrip_chars_i(xs_str *str, const char *chars)
+/* strips all chars from the start of str */
+{
+    int n;
+
+    for (n = 0; str[n] && strchr(chars, str[n]); n++);
+
+    if (n)
+        str = xs_collapse(str, 0, n);
+
+    return str;
+}
+
+
+xs_str *xs_rstrip_chars_i(xs_str *str, const char *chars)
+/* strips all chars from the end of str */
 {
-    XS_ASSERT_TYPE(str, XSTYPE_STRING);
-
     int n;
 
-    /* strip first from the end */
     for (n = strlen(str); n > 0 && strchr(chars, str[n - 1]); n--);
     str[n] = '\0';
 
-    if (str[0]) {
-        /* now strip from the beginning */
-        for (n = 0; str[n] && strchr(chars, str[n]); n++);
-
-        if (n)
-            str = xs_collapse(str, 0, n);
-    }
-
     return str;
 }
 
 
+xs_str *xs_strip_chars_i(xs_str *str, const char *chars)
+/* strips the string of chars from the start and the end */
+{
+    return xs_lstrip_chars_i(xs_rstrip_chars_i(str, chars), chars);
+}
+
+
 xs_str *xs_tolower_i(xs_str *str)
 /* convert to lowercase */
 {
@@ -603,6 +626,11 @@ xs_list *_xs_list_write_litem(xs_list *list, int offset, const char *mem, int ds
 {
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
+    if (mem == NULL) {
+        mem = xs_stock_null;
+        dsz = sizeof(xs_stock_null);
+    }
+
     list = xs_expand(list, offset, dsz + 1);
 
     list[offset] = XSTYPE_LITEM;
@@ -823,17 +851,18 @@ xs_str *xs_join(const xs_list *list, const char *sep)
             int sz;
 
             /* add the separator */
-            if (c != 0) {
+            if (c != 0 && ssz) {
                 s = xs_realloc(s, offset + ssz);
                 memcpy(s + offset, sep, ssz);
                 offset += ssz;
             }
 
             /* add the element */
-            sz = strlen(v);
-            s = xs_realloc(s, offset + sz);
-            memcpy(s + offset, v, sz);
-            offset += sz;
+            if ((sz = strlen(v)) > 0) {
+                s = xs_realloc(s, offset + sz);
+                memcpy(s + offset, v, sz);
+                offset += sz;
+            }
 
             c++;
         }
@@ -857,11 +886,9 @@ xs_list *xs_split_n(const char *str, const char *sep, int times)
     list = xs_list_new();
 
     while (times > 0 && (ss = strstr(str, sep)) != NULL) {
-        /* add the first part (without the asciiz) */
-        list = xs_list_append_m(list, str, ss - str);
-
-        /* add the asciiz */
-        list = xs_insert_m(list, xs_size(list) - 1, "", 1);
+        /* create a new string with this slice and add it to the list */
+        xs *s = xs_str_new_sz(str, ss - str);
+        list = xs_list_append(list, s);
 
         /* skip past the separator */
         str = ss + sz;
@@ -899,13 +926,18 @@ xs_dict *xs_dict_new(void)
 }
 
 
-xs_dict *xs_dict_insert_m(xs_dict *dict, int offset, const xs_str *key,
+xs_dict *_xs_dict_write_ditem(xs_dict *dict, int offset, const xs_str *key,
                           const xs_val *data, int dsz)
 /* inserts a memory block into the dict */
 {
     XS_ASSERT_TYPE(dict, XSTYPE_DICT);
     XS_ASSERT_TYPE(key, XSTYPE_STRING);
 
+    if (data == NULL) {
+        data = xs_stock_null;
+        dsz = sizeof(xs_stock_null);
+    }
+
     int ksz = xs_size(key);
 
     dict = xs_expand(dict, offset, 1 + ksz + dsz);
@@ -921,14 +953,14 @@ xs_dict *xs_dict_insert_m(xs_dict *dict, int offset, const xs_str *key,
 xs_dict *xs_dict_append_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz)
 /* appends a memory block to the dict */
 {
-    return xs_dict_insert_m(dict, xs_size(dict) - 1, key, mem, dsz);
+    return _xs_dict_write_ditem(dict, xs_size(dict) - 1, key, mem, dsz);
 }
 
 
 xs_dict *xs_dict_prepend_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz)
 /* prepends a memory block to the dict */
 {
-    return xs_dict_insert_m(dict, 4, key, mem, dsz);
+    return _xs_dict_write_ditem(dict, 4, key, mem, dsz);
 }
 
 
@@ -1121,11 +1153,10 @@ int xs_data_size(const xs_data *value)
 }
 
 
-void xs_data_get(const xs_data *value, void *data)
+void xs_data_get(void *data, const xs_data *value)
 /* copies the raw data stored inside value into data */
 {
-    int size = _xs_get_24b(value + 1) - 4;
-    memcpy(data, &value[4], size);
+    memcpy(data, &value[4], xs_data_size(value));
 }
 
 
